from FX import *
import Skybox

theta = 0.9
enableRefraction = 1.0
enableFresnel = 1.0

amplification = 2.0
offset = 0.4

inpurityTex = Texture('star_impuritymap.png', TEXTURE_2D)
randomTex = Texture('randomTexture.png', TEXTURE_2D)


vs = Cg("""
// fresnel approximation
fixed fast_fresnel(float3 I, float3 N, float3 fresnelValues)
{
    fixed power = fresnelValues.x;
    fixed scale = fresnelValues.y;
    fixed bias = fresnelValues.z;

    return bias + pow(1.0 - dot(I, N), power) * scale;
}

void main(float4 position     : POSITION,
          float4 normal       : NORMAL,
          float2 parameter    : TEXCOORD0,
 
      out float4 hPosition    : POSITION,
      out float  fresnelTerm  : COLOR0,
      out float2 texCoord     : TEXCOORD0,
      out float4 refractVec   : TEXCOORD1,
      out float4 reflectVec   : TEXCOORD2,
         
  uniform float4x4 ModelViewProj,
  uniform float4x4 ModelView,
  uniform float4x4 ModelViewIT,
  uniform float    theta)
{
    texCoord = parameter;
    hPosition = mul(ModelViewProj, position);

    // convert the position and normal into appropriate spaces
    float3 eyePosition = mul(ModelView, position).xyz;
    float3 eyeToVert = normalize(eyePosition);
        
    float3 eyeNormal = mul(ModelViewIT, normal).xyz;
    eyeNormal = normalize(eyeNormal);

    // calculate refraction
    refractVec.xyz = refract(eyeToVert, eyeNormal, theta);
    refractVec.w = 0;

    // calculate reflection
    reflectVec.xyz = reflect(eyeToVert, eyeNormal);
    reflectVec.w = 0;

    // calculate fresnel term
    fresnelTerm = fast_fresnel(-eyeToVert, eyeNormal, 
                               float3(2.0, 2.0, 0.1));
}""")

fs = Cg("""
/**
 * This function creates a cosine powered distributed vector around
 * the supplied center vector. The power and the random numbers
 * xis are used to determine the returned vector. These calculations
 * are computationally expensive and consumes a lot of instructions.
 */
float3 cosine_powered(float3 center, float power, float2 xis) {
    float pi = 3.1415;
    float phi = 2 * pi * xis[0];
    float theta = acos(pow(xis[1], 1 / (power + 1)));
    float x = cos(2 * pi * xis[0]) * sqrt(1 - pow(xis[1], 2 / (power + 1)));
    float y = x;
    float z = pow(xis[1], 1 / (power + 1));

    float3 randomDirection = float3(x, y, z);

    // Construct a new matrix, initially the unit matrix.
    float3x3 rotate = float3x3(1, 0, 0, 0, 1, 0, 0, 0, 1);
    x = center.x;
    y = center.y;
    z = center.z;
    float x2y2 = x * x + y * y;
    
    if (x2y2 > 1e-6) {
        // Set first column.
        rotate._m00 = (x * x * z + y * y) / x2y2;
        rotate._m10 = (x * y * (z - 1)) / x2y2;
        rotate._m20 = -x;
        
        // Set second column.
        rotate._m01 = (x * y * (z - 1)) / x2y2;
        rotate._m11 = (y * y * (z - 1)) / x2y2 + 1;
        rotate._m21 = -y;
        
        // Set third column.
        rotate._m02 = x;
        rotate._m12 = y;
        rotate._m22 = z;
    }

    return mul(rotate, randomDirection);
}


void main(float2 texCoord   : TEXCOORD0,
          float3 refractVec : TEXCOORD1,
          float3 reflectVec : TEXCOORD2,
          float fresnelTerm : COLOR0,

      out float3 color : COLOR0,

  uniform samplerCUBE environmentMaps[2],
  uniform float       enableRefraction,
  uniform float       enableFresnel,
  uniform float       offset,
  uniform float       amplification,
  uniform sampler2D   inpurityMap,
  uniform sampler2D   randomMap)
{

    // Calculates the inpurity factor. Lower values are more inpure and
    // creates larger variations in the reflected and refracted rays.
    // The inpurity scale is exponential with the base 1000. Inpurity
    // is controlled through an offset and an amplification factor. These
    // factors creates the proper balance between black and white in the
    // inpurity map.
    float inpurity = pow(1000, 
                         (tex2D(inpurityMap, texCoord).r + offset) * 
                         amplification);

    // Look up random numbers in the uniformly distributed random number map.
    float2 xis = tex2D(randomMap, texCoord).rg;

    // Perturbation of the reflection and refraction vectors.
    refractVec = cosine_powered(refractVec, inpurity, xis);
    //reflectVec = cosine_powered(reflectVec, inpurity, xis);

    float3 refractColor = texCUBE(environmentMaps[0], refractVec).rgb; 
    float3 reflectColor = texCUBE(environmentMaps[1], reflectVec).rgb; 
    float3 reflectRefract = lerp(refractColor, reflectColor, fresnelTerm);
    
    // The final color depends on if fresnel and refraction is
    // is enabled or not. With everything enabled the resulting
    // color is a weighted value between the reflected and the refracted
    // term.
    color = (enableRefraction) ? 
        ((enableFresnel) ? (reflectRefract) : (refractColor)) : 
        ((enableFresnel) ? (reflectColor) : fresnelTerm);
}""")

environmentMaps = [Skybox.cubeMap, Skybox.cubeMap]

randomMap = Sampler(texture = randomTex,
                    MinFilter = GL_LINEAR,
                    MagFilter = GL_LINEAR)

inpurityMap = Sampler(texture = inpurityTex,
                      MinFilter = GL_LINEAR,
                      MagFilter = GL_LINEAR)


technique = [Pass(ZEnable = True,
                  VertexShader = vs(target = arbvp1),
                  FragmentShader = fs(target = arbfp1))]


